
from django.test import TestCase

from django.urls import reverse


class PageDecoratorTest(TestCase):

    def test_view_returns_template(self):
        response = self.client.get(
            reverse('template_view'),
        )
        self.assertTemplateUsed(
            response,
            'simple_template.html'
        )



from __future__ import unicode_literals

from django import VERSION
from django.contrib import admin
from django.contrib.admin.models import ADDITION, CHANGE, DELETION, LogEntry
from django.contrib.admin.utils import construct_change_message
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.forms import all_valid
from django.template.defaultfilters import capfirst
from django.utils.safestring import mark_safe
from django.utils.text import get_text_list

from .readonly_admin import ReadOnlyModelAdmin

# Get rid off warnings in Django 3
if VERSION[0] >= 2:
    from django.utils.encoding import force_str as force_text
    from django.utils.translation import gettext_lazy as _

    LogEntry.get_change_message.short_description = _("Changes in detail")
else:
    # @RemoveFromDjangoVersion2
    from django.utils.encoding import force_text
    from django.utils.translation import ugettext_lazy as _


UserModel = get_user_model()


class LogEntryAs(object):
    """
    This creates the LogEntry items for actions on objects.
    Similar to how ModelAdmin does it. This class is more universal to embrace various scenarios.

    Replaces the old-fashioned UILogEntry below.
    """
    def __init__(self, user, **common_message):
        """
        Expects the actioning user as argument.
        Additional parameters will be saved each time to JSON message each time the action occurs.
        """
        self.user = user
        self.common_message_list = [common_message]

    def log_addition(self, object_, message=None):
        """
        Adds a LogEntry recording that the object_ was added.
        The message is either a dict which is appended to the resulting message list.
        Otherwise, it is expected to be an iterable and the resulting message list is extended with it.
        """
        self.log_action(object_, ADDITION, message)

    def log_change(self, object_, message=None):
        """
        Adds a LogEntry recording that the object_ was changed.
        The message is either a dict which is appended to the resulting message list.
        Otherwise, it is expected to be an iterable and the resulting message list is extended with it.
        """
        self.log_action(object_, CHANGE, message)

    def log_deletion(self, object_, message=None):
        """
        Adds a LogEntry recording that the object_ was deleted.
        NOTE that this method must be called before the deletion.
        The message is either a dict which is appended to the resulting message list.
        Otherwise, it is expected to be an iterable and the resulting message list is extended with it.
        """
        self.log_action(object_, DELETION, message)

    def log_action(self, object_, action_flag, message):
        change_message = []
        if self.common_message_list:
            change_message.extend(self.common_message_list)
        if message:
            if isinstance(message, dict):
                change_message.append([message])
            else:
                change_message.extend(message)

        LogEntry.objects.log_action(
            user_id=self.user.pk,
            content_type_id=ContentType.objects.get_for_model(object_, for_concrete_model=False).pk,
            object_id=object_.pk,
            object_repr=force_text(object_),
            action_flag=action_flag,
            change_message=change_message,
        )

    def log_form_addition_or_change(self, form, was_adding, formsets=None):
        """
        Shortcut for logging the addition/deletion made via form. Resembles ModelAdmin too.
        If there are accompanyed formset to the main action, their change could be logged too.
        """
        if not form.is_valid():
            raise ValidationError("Form invalid.")
        if formsets and not all_valid(formsets):
            raise ValidationError("Some of the formsets not valid.")

        message_list = construct_change_message(form, formsets, was_adding)

        if was_adding:
            self.log_addition(form.instance, message_list)
        else:
            self.log_change(form.instance, message_list)


class UILogEntry(object):
    """
    Deprecated, you should migrate to LogEntryAs above.
    """
    def __init__(self, user=None):
        """
        This is similar to how ModelAdmin logs events. Expects the user
        doing the change as argument. When not given, 'admin' is used.
        """
        if user is None:
            user = UserModel.objects.only('id').get(username='admin')
        self.user = user

    def log_addition(self, object_, message='UI'):
        """
        Adds a LogEntry recording that the object_ was added.
        The change detail message defaults to 'UI' to denote the change
        did not originate from admin interface.
        """
        self.do_log(object_, ADDITION, message)

    def log_change(self, object_, message='UI', form=None):
        """
        Adds a LogEntry recording that the object_ was changed.
        If clean form given, the list of changed fields is added to message.
        The change detail message defaults to 'UI' to denote the change
        did not originate from admin interface.
        """
        if form is not None:
            if message:
                message += ' '
            message += self.construct_change_message(form)

        self.do_log(object_, CHANGE, message)

    def log_deletion(self, object_, message='UI'):
        """
        Adds a LogEntry recording that the object_ was deleted.
        The change detail message defaults to 'UI' to denote the change
        did not originate from admin interface.
        """
        self.do_log(object_, DELETION, message)

    @staticmethod
    def construct_change_message(form):
        change_message = []
        if form.changed_data:
            lst = get_text_list(form.changed_data, _('and'))
            change_message.append(_('Changed %s.') % lst)

        change_message = ' '.join(change_message)
        return force_text(change_message or _('No fields changed.'))

    def do_log(self, object_, action_flag, message):
        LogEntry.objects.log_action(
            user_id=self.user.pk,
            content_type_id=ContentType.objects.get_for_model(object_).pk,
            object_id=object_.pk,
            object_repr=force_text(object_),
            action_flag=action_flag,
            change_message=force_text(message),
        )

    def save_form_instance_and_log(self, form):
        """
        Shortcut to saving a clean ModelForm instance + logging
        the addition or deletion.
        """
        existed = form.instance.id

        instance = form.save()

        if existed:
            self.log_change(form.instance, form=form)
        else:
            self.log_addition(form.instance)

        return instance


class LogEntryAdmin(ReadOnlyModelAdmin):
    """
    This read only ModelAdmin can be presented to the staff users for reference.

    Example::

        from django.contrib import admin
        from django.contrib.admin.models import LogEntry
        from useful.django.logentry import LogEntryAdmin

        admin.site.register(LogEntry, LogEntryAdmin)
    """
    list_display = (
        'action_time_nolink',
        'user',
        'table',
        'action',
        'get_change_message',
    )
    list_filter = (
        'action_flag',
        'user__is_superuser',
        ('content_type', admin.RelatedOnlyFieldListFilter),
    )
    list_select_related = (
        'user',
        'content_type',
    )
    search_fields = (
        'user__email',
        'object_repr',
        'change_message',
    )
    date_hierarchy = 'action_time'
    list_per_page = 500
    list_max_show_all = 10000

    def action_time_nolink(self, obj):
        return mark_safe(
            '</a><span style="font-weight: normal">%s</span><a>' % obj.action_time.strftime('%Y-%m-%d %H:%M:%S')
        )
    action_time_nolink.short_description = _("Action time")
    action_time_nolink.admin_order_field = 'action_time'

    def table(self, obj):
        if obj.content_type_id:
            cls = obj.content_type.model_class()
            return capfirst(cls._meta.verbose_name_plural) if cls else obj.content_type
        return ''
    table.short_description = _("Table")
    table.admin_order_field = 'content_type'

    def action(self, obj):
        r = obj.object_repr
        if not obj.is_deletion() and obj.get_admin_url():
            r = '<a href="%s">%s</a>' % (
                obj.get_admin_url(),
                r,
            )
        return mark_safe(
            '<span class="%s">%s</span>' % (
                self.link_classes[obj.action_flag],
                r,
            )
        )
    action.short_description = _("Action, object")
    action.admin_order_field = 'object_repr'

    link_classes = {
        admin.models.ADDITION: 'addlink',
        admin.models.CHANGE: 'changelink',
        admin.models.DELETION: 'deletelink',
    }



from django.test import TestCase
from django.utils.crypto import get_random_string

from tests.testapp.utils import expensive_function


# @override_settings(AUTHENTICATION_BACKENDS=['useful.django.auth.EmailLoginModelBackend', ])
class CachedFunctionTest(TestCase):

    def test_returns_result(self):
        # counts provided arguments
        result = expensive_function(
            1, 'foo', None, 3.123, True,
        )
        self.assertEqual(result, 5)

    def test_returns_result_for_long_arguments(self):
        result = expensive_function(
            1, 'foo', None, 3.123, True,
            get_random_string(200),
            get_random_string(300),
        )
        self.assertEqual(result, 7)

    def test_bytestring(self):
        b_text = b'\xc4\xbe\xc5\xa1\xc4\x8d\xc5\xa5\xc5\xbe\xc3\xbd\xc3\xa1\xc3\xad\xc3\xa9' * 10
        result = expensive_function(
            b_text, b_text.decode('utf-8')
        )
        self.assertEqual(result, 2)



from django.http import HttpResponse
from django.shortcuts import render

from useful.django.urlpatterns import UrlPatterns
from useful.django.views import page

app_name = 'testapp'
urlpatterns = UrlPatterns()


@urlpatterns.url(r'^url-simple-view/$')
def simple_view(request):
    return HttpResponse('View works!')


@urlpatterns.url(r'^other-view/$')
def other_view(request):
    return HttpResponse('Other view works!')


@page('simple_template.html')
def template_view(request):
    return dict(
        foo='bar',
    )


@urlpatterns.url(r'^restricted/$', perms='auth.can_pass')
def restricted_view(request):
    return render(
        request,
        'restricted_template.html',
    )


@urlpatterns.url(r'^restricted2/$', perms='auth.other_perm')
def another_restricted_view(request):
    return render(
        request,
        'restricted_template.html',
    )




from django.contrib.auth import authenticate
from django.contrib.auth.models import User
from django.test import TestCase
from django.test import override_settings


@override_settings(AUTHENTICATION_BACKENDS=['useful.django.auth.EmailLoginModelBackend', ])
class EmailLoginModelBackendTest(TestCase):

    def test_authenticates_using_email(self):
        user = User(
            username='foobar',
            email='foo@bar.baz'
        )
        user.set_password('qwerty321')
        user.save()

        auth_user = authenticate(username='foo@bar.baz', password='qwerty321')
        self.assertEqual(auth_user, user)

        auth_user = authenticate(username='foobar', password='qwerty321')
        self.assertIsNone(auth_user)


@override_settings(AUTHENTICATION_BACKENDS=['useful.django.auth.UsernameOrEmailLoginModelBackend', ])
class UsernameOrEmailLoginModelBackendTest(TestCase):

    def test_authenticates_using_email(self):
        user = User(
            username='foobar',
            email='foo@bar.baz'
        )
        user.set_password('qwerty321')
        user.save()

        auth_user = authenticate(username='foo@bar.baz', password='qwerty321')
        self.assertEqual(auth_user, user)

        auth_user = authenticate(username='foobar', password='qwerty321')
        self.assertEqual(auth_user, user)



from django.http import QueryDict
from django.template.loader import render_to_string
from django.test import TestCase


class TemplateTagsTest(TestCase):

    def test_useful_tags(self):
        body_lines = render_to_string(
            'template_for_tags.html',
            context=dict(
                querydict=QueryDict('a=12&b=13'),
                map1={
                    'key1': 'KEY',
                },
                int1=300,
                int2=3000,
                int3=30000,
                int4=300000,
                int1m=-300,
                int2m=-3000,
                int3m=-30000,
                int4m=-300000,
                string1='test',
                string2='TEST',
                begin='te',
                path='/etc/passwd/shadow',
            )
        ).strip().split('\n')

        expected_lines = (
            ('a=newvalue&amp;b=13', 'b=13&amp;a=newvalue'),  # result should be one of these
            'KEY',
            'KEY',
            '300',
            '3 000',
            '30 000',
            '300 000',
            '-300',
            '-3 000',
            '-30 000',
            '-300 000',
            'True',
            'False',
            'shadow',
        )

        for bl, el in zip(body_lines, expected_lines):
            if isinstance(el, tuple):
                self.assertIn(bl, el)
            else:
                self.assertEqual(bl, el)

        self.assertEqual(len(body_lines), len(expected_lines))



from django.conf.global_settings import *


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "mem_db"
    }
}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "useful.django",
    "tests.testapp.apps.TestAppConfig"
]

SECRET_KEY = 'justfortests'

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
    }
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    }
}



import json

from django import VERSION
from django.contrib.admin.models import LogEntry
from django.contrib.auth.models import Group, User
from django.forms import ModelForm
from django.http import QueryDict
from django.test import TestCase

from useful.django.logentry import LogEntryAs


class GroupModelForm(ModelForm):
    class Meta:
        model = Group
        fields = 'name',


class LogEntryAsTest(TestCase):

    def setUp(self):
        user = User.objects.create(username='somebody')
        db_logger = LogEntryAs(user, ui=True)

        new_group = Group.objects.create(name='group1')
        db_logger.log_addition(new_group)

        group = Group.objects.get(name='group1')
        form = GroupModelForm(
            data=QueryDict('name=group2'),
            instance=group,
        )
        self.assertTrue(form.is_valid())

        form.save()

        db_logger.log_form_addition_or_change(
            form,
            was_adding=False,
        )

        group = Group.objects.get(name='group2')
        db_logger.log_deletion(group)
        # Not really deleting. :-)

    def test_log_entry(self):
        user = User.objects.get(username='somebody')
        group = Group.objects.get(name='group2')

        self.assertEqual(LogEntry.objects.count(), 3)

        entry1, entry2, entry3 = list(
            LogEntry.objects.order_by('id')
        )

        # Since Django 3.0 the Admin's model history change messages
        # now prefers more readable field labels instead of field names.
        field_desc = "Name" if VERSION[0] >= 3 else "name"
        dash = "\u2014" if VERSION[0] >= 3 else "-"

        self.assertEqual(entry1.user, user)
        self.assertEqual(entry1.object_repr, 'group1')
        self.assertTrue(entry1.is_addition())
        self.assertEqual(
            entry1.change_message,
            json.dumps(
                [{'ui': True}]
            )
        )
        self.assertEqual(str(entry1), 'Added %s.' % s_quote("group1"))
        self.assertEqual(
            entry1.get_change_message(),
            "No fields changed.",
        )
        self.assertEqual(
            entry1.get_edited_object(), group,
        )

        self.assertEqual(entry2.user, user)
        self.assertEqual(entry2.object_repr, 'group2')
        self.assertTrue(entry2.is_change())
        self.assertEqual(
            entry2.change_message,
            json.dumps(
                [{'ui': True}, {'changed': {'fields': [field_desc]}}]
            )
        )
        self.assertEqual(
            str(entry2),
            'Changed %s %s Changed %s.' % (s_quote("group2"), dash, field_desc)
        )
        self.assertEqual(
            entry2.get_change_message(),
            "Changed %s." % field_desc,
        )
        self.assertEqual(
            entry2.get_edited_object(), group,
        )

        self.assertEqual(entry3.user, user)
        self.assertEqual(entry3.object_repr, 'group2')
        self.assertTrue(entry3.is_deletion())
        self.assertEqual(
            entry3.change_message,
            json.dumps(
                [{'ui': True}]
            )
        )
        # Django bug https://code.djangoproject.com/ticket/32136
        self.assertEqual(str(entry3), 'Deleted %s' % s_quote("group2."))
        self.assertEqual(
            entry3.get_change_message(),
            "No fields changed.",
        )
        self.assertEqual(
            entry3.get_edited_object(), group,
        )

        group.delete()


if VERSION[0] >= 3:
    def s_quote(s):
        return '\u201C%s\u201D' % s
else:
    def s_quote(s):
        return '"%s"' % s




import sys


PY2 = sys.version_info[0] == 2


class Choices:
    """
    Encapsulates the choices definition and conversion. Praises the DRY.

    Allows for using custom symbols as individual choices, bundles them with i18n-enabled labels
    and stores choices in a way that usually fits more for processing, such as small integers.

    Example::

        class Payment(models.Model):
            STATUSES = Choices(
                initial      = (100, _("Awaiting Payment")),
                success      = (200, _("Paid OK")),
                nok_failure  = (300, _("Payment Failed")),
            )
            ...
            STATUSES = models.SmallIntegerField(
                ...,
                choices=STATUSES.choices(),
                default=STATUSES.initial,
            )
            ...

    Second choice can be referred to as Payment.STATUSES.success throughout the code, equals to 200.
    """
    reserved_choice_names = {
        'contribute_to_class',
    }

    def __init__(self, **defs):
        """
        Takes the dictionary, where keys are identifiers and values are tuples (db id, text).
        """
        if any(k.startswith('_') for k in defs):
            raise ValueError("Names started by _ are not allowed in Choices.")

        unallowed = set(defs) & self.reserved_choice_names
        if unallowed:
            raise ValueError("Not allowed choice names defined: %s" % ', '.join(unallowed))

        self.defs = defs
        self.names = {v[0]: k for k, v in defs.items()}
        self.texts = dict(defs.values())

        if len(set(self.names)) != len(defs):
            raise ValueError("Duplicate choices defined.")

    def __add__(self, other):
        """
        Support for Choices(...) + Choices(...) by shallow copy.
        """
        defs = self.defs.copy()
        defs.update(other.defs)
        return Choices(**defs)

    def choices(self):
        ch = list(
            self.defs.values()
        )
        ch.sort(
            key=lambda x: x[0],
        )
        return ch

    def get(self, name, default=None):
        """
        Example: Choices(...).get('success') -> 200
        """
        return self.defs.get(
            name,
            (default,)
        )[0]

    def __getattr__(self, name):
        """
        Example: Choices(...).success -> 200
        """
        if name.startswith('_') or name in self.reserved_choice_names:
            raise AttributeError("No such attribute %r in Choices class." % name)

        return self.defs[name][0]

    def name_of(self, what):
        """
        Example: name_of(200) -> 'success'
        """
        return self.names[what]

    def text_of(self, what):
        """
        Example: text_of(200) -> "Paid OK"
        """
        return self.texts[what]

    def tuples(self):
        """
        Generates tuples (db id, short name, text).
        """
        for k, v in self.defs.items():
            # Enforcing i18n by unicode/str()
            if PY2:
                # noinspection PyUnresolvedReferences
                yield v[0], k, unicode(v[1])
            else:
                yield v[0], k, str(v[1])

    def select(self, keys):
        """
        Returns new Choice object with subset of choices.
        """
        if not isinstance(keys, (list, tuple)):
            keys = [keys]

        return Choices(
            **dict((k, self.defs[k])
                   for k in keys if k in self.defs)
        )




import sys

from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.test import override_settings
from django.urls import reverse

PY2 = sys.version_info[0] == 2

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'useful.django.can_url.can_url_processor',
            ]
        }
    }
]


def test_resolves_view_defined_url():
    result = reverse('testapp:simple_view')
    assert result == '/testapp/url-simple-view/'

    result = reverse('testapp:other_view')
    assert result == '/testapp/other-view/'


@override_settings(TEMPLATES=TEMPLATES)
class CanUrlTest(TestCase):

    def setUp(self):
        super(CanUrlTest, self).setUp()
        content_type = ContentType.objects.get_for_model(User)
        self.permission = Permission.objects.create(
            codename='can_pass',
            name='Can pass',
            content_type=content_type,
        )
        self.user = User.objects.create(
            username='foo',
        )

    def test_redirects_to_login_if_no_perm(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('testapp:restricted_view'),
        )
        self.assertRedirects(
            response,
            '%s?next=%s' % (
                settings.LOGIN_URL,
                reverse('testapp:restricted_view')
            ),
            fetch_redirect_response=False
        )

    def test_returns_rendered_template_if_perm_ok(self):
        self.user.user_permissions.add(self.permission)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('testapp:restricted_view'),
        )
        content = response.content
        if not PY2:
            content = content.decode('utf-8')

        print("XXX")
        print(type(content))
        self.assertEqual(response.status_code, 200)
        self.assertIn('restricted_view', content)
        self.assertNotIn('another_restricted_view', content)




from django.conf.urls import include
from django.conf.urls import url

from tests.testapp import views


urlpatterns = [
    url(r'testapp/', include('tests.testapp.views')),
    url(
        r'testapp/simple/',
        views.template_view,
        name='template_view'
    )
]



from django.contrib.auth.models import Group, User
from django.test import TestCase

from useful.django.getters import prefetch_m2m


class PrefetchM2MTest(TestCase):

    def test_returns_dict_of_lists(self):
        user1 = User.objects.create(
            username='foobar',
            email='foo@bar.baz'
        )
        user2 = User.objects.create(
            username='foobar2',
            email='foo2@bar.baz'
        )
        group1 = Group.objects.create(name='Foo')
        group2 = Group.objects.create(name='Bar')
        group3 = Group.objects.create(name='Baz')
        user1.groups.set([group1, group2])
        user2.groups.set([group2, group3])

        groups_m2m = prefetch_m2m(User.groups)

        self.assertEqual(
            list(groups_m2m.keys()),
            [user1.id, user2.id]
        )
        self.assertEqual(groups_m2m[user1.id], [group1, group2])
        self.assertEqual(groups_m2m[user2.id], [group2, group3])
